import peertable as peer
import threading
import json
import random
import string
import wx
import time


class ChatApp(peer.PeerApplication):
    def __init__(self, nickname):
        self.nickname = nickname
        self.messages = []
        self.nicknames = {}
        self.users = []
        self.server = None

    def on_app_message(self):
        pass
        
    def send_message(self, msg):
        if self.server is None:
            return
    
        mid = ''.join([random.choice(string.ascii_letters + string.digits) for _ in range(50)])
        
        self.messages.append(("MESSAGE", time.time(), self.server.id, ))
        self.on_app_message()
        
        self.server.broadcast(peer.Message(True, "MESSAGE", json.dumps({
            'sender_id': self.server.id,
            'message_id': mid,
            'content': msg,
            'closed': [self.server.id]
        })))
        
    def quit(self):
        if self.server is None:
            return
            
        self.server.broadcast(peer.Message(True, "QUIT", "{}:{}:{}".format(self.server.id, ''.join([random.choice(string.ascii_letters + string.digits) for _ in range(50)]), self.server.id)))
        
    def set_nick(self, nick):
        if self.server is None:
            return
    
        self.nickname = nick
        self.nicknames[self.server.id] = nick
        self.server.broadcast(peer.Message(True, "MYNAMEIS", "{}:{}:{}".format(self.server.id, ''.join([random.choice(string.ascii_letters + string.digits) for _ in range(50)]), self.server.id)))
        
    def receive(self, server, client, message):
        if message.message_type == "GETNICK ":
            server.send_id(client.id, peer.Message(True, "MYNAMEIS", self.nickname))
            client.disconnect() # here to prevent that too many clients populate our lists
            
        elif message.message_type == "MYNAMEIS":
            self.nicknames[client.id] = message.payload.encode('utf-8')
            
            if client.id not in self.users:
                self.users.append(client.id)
                
        elif message.message_type == "QUIT    ":
            quitter, quit_id, closed = message.payload.decode('utf-8').split(':')
            closed = closed.split('#')
            
            if quit_id in [x[3] for x in self.messages if x[0] == "QUIT"]:
                return
            
            self.messages.append(("QUIT", time.time(), quit_id, quitter))
            self.on_app_message()
            
            for c in server.clients:
                if c.id == quitter:
                    c.disconnect()
            
                elif c.id not in closed:
                    closed.append(c.id)
                    
            server.broadcast(peer.Message(True, "QUIT", "{}:{}:{}".format(quitter, quit_id, '#'.join(closed))))
    
        elif message.message_type == "JOIN    ":
            joiner, join_id, closed = message.payload.decode('utf-8').split(':')
            closed = closed.split('#')
            
            if join_id in [x[3] for x in self.messages if x[0] == "JOIN"] or joiner == self.server.id:
                return
            
            self.messages.append(("JOIN", time.time(), join_id, joiner))
            self.on_app_message()
            
            for c in server.clients:
                if c.id == joiner:
                    c.disconnect()
            
                elif c.id not in closed and c.id is not None:
                    closed.append(c.id)
                    
            server.broadcast(peer.Message(True, "JOIN", "{}:{}:{}".format(joiner, join_id, '#'.join(closed))))
    
        elif message.message_type == "__QUERYR":
            cid = message.payload.decode('utf-8')
            server.send_id(cid, peer.Message(True, "GETNICK"))
            
        elif message.message_type == "MESSAGE ":
            data = json.loads(message.payload.decode('utf-8'))
            
            for m in self.messages:
                if m[0] == "MESSAGE" and m[2] == data['message_id']:
                    return
            
            self.messages.append(("MESSAGE", time.time(), data['sender_id'], data['message_id'], data['content']))
            self.on_app_message()
            
            for c in server.clients:
                if c.id not in data['closed']:
                    data['closed'].append(c.id)
                    
            server.broadcast(peer.Message(True, "MESSAGE", json.dumps(data).encode('utf-8')))
        
    def on_registered(self, server):
        self.server = server
        self.nicknames[self.server.id] = self.nickname
        server.broadcast(peer.Message(False, "QUERYALL", "{}:{}:{}:{}:{}".format(*server.address, server.id, ''.join([random.choice(string.digits + string.ascii_letters) for _ in range(40)]), '#'.join([c.id for c in server.clients if c and c.id] + [server.id]))))
        
    def connected(self, server, client):   
        # wait until client has identified
        def update():
            while client.id is None:
                time.sleep(0.1)
        
            jid = ''.join([random.choice(string.ascii_letters + string.digits) for _ in range(50)])
            self.messages.append(("JOIN", time.time(), jid, client.id))
            self.on_app_message()
            
            server.broadcast(peer.Message(True, 'JOIN', "{}:{}:{}".format(client.id, jid, self.server.id)))
            client.send(peer.Message(True, "GETNICK"))
            
        threading.Thread(target=update).start()


import yaml

from setuptools import setup, find_packages
from os import path

# read the contents of README file
this_dir = path.abspath(path.dirname(__file__))

with open(path.join(this_dir, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="Peertalk",
    version=yaml.load(open("changelog.yml"))['versions'][-1]['name'],
    packages=["peertalk"],
    
    # metadata to display on PyPI
    author="Gustavo6046",
    author_email="gugurehermann@gmail.com",
    description="An experimental, demonstrative chat protocol and implementation for use in Peertable networks",
    license="MIT",
    keywords=" ".join(['p2p', 'peer-to-peer', 'peer', 'network', 'infrastructure', 'chat', 'test', 'experimental', 'protocol', 'demo', 'talk']),
    
    long_description=long_description,
    long_description_content_type='text/markdown'
)


import numpy as np
import threading
import os
import json
import random
import string
import wx
import peertalk as pt
import peertable as pb


class InitDialog(wx.Dialog):
    def __init__(self, parent, container, app):
        wx.Dialog.__init__ ( self, parent, id = wx.ID_ANY, title = u"First-Time Setup", pos = wx.DefaultPosition, size = wx.Size( 559,217 ), style = wx.DEFAULT_DIALOG_STYLE )
        self.SetSizeHints( wx.DefaultSize, wx.DefaultSize )
        bSizer1 = wx.BoxSizer( wx.VERTICAL )
        self.m_staticText1 = wx.StaticText( self, wx.ID_ANY, u"Please fill in the fields below so you can receive data, have an online pseudonym, et cetera.", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_staticText1.Wrap( -1 )
        bSizer1.Add( self.m_staticText1, 0, wx.ALL, 5 )
        self.m_staticline1 = wx.StaticLine( self, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, wx.LI_HORIZONTAL )
        bSizer1.Add( self.m_staticline1, 0, wx.EXPAND |wx.ALL, 5 )
        bSizer2 = wx.BoxSizer( wx.HORIZONTAL )
        self.m_staticText2 = wx.StaticText( self, wx.ID_ANY, u"Nickname:", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_staticText2.Wrap( -1 )
        bSizer2.Add( self.m_staticText2, 0, wx.ALL, 5 )
        self.startnick = wx.TextCtrl( self, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, 0 )
        bSizer2.Add( self.startnick, 1, wx.ALL, 5 )
        bSizer1.Add( bSizer2, 0, wx.EXPAND, 5 )
        bSizer3 = wx.BoxSizer( wx.HORIZONTAL )
        self.m_staticText3 = wx.StaticText( self, wx.ID_ANY, u"External address (to receive data):", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_staticText3.Wrap( -1 )
        bSizer3.Add( self.m_staticText3, 0, wx.ALL, 5 )
        self.remoteaddr = wx.TextCtrl( self, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, 0 )
        bSizer3.Add( self.remoteaddr, 1, wx.ALL, 5 )
        bSizer1.Add( bSizer3, 0, wx.EXPAND, 5 )
        bSizer31 = wx.BoxSizer( wx.HORIZONTAL )
        self.m_staticText31 = wx.StaticText( self, wx.ID_ANY, u"Peer listen port:", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_staticText31.Wrap( -1 )
        bSizer31.Add( self.m_staticText31, 0, wx.ALL, 5 )
        self.m_spinCtrl1 = wx.SpinCtrl( self, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, wx.SP_ARROW_KEYS, 2800, 2999, 2912 )
        bSizer31.Add( self.m_spinCtrl1, 1, wx.ALL, 5 )
        bSizer1.Add( bSizer31, 0, wx.EXPAND, 5 )
        self.m_staticline2 = wx.StaticLine( self, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, wx.LI_HORIZONTAL )
        bSizer1.Add( self.m_staticline2, 0, wx.EXPAND |wx.ALL, 5 )
        bSizer9 = wx.BoxSizer( wx.HORIZONTAL )
        self.m_button1 = wx.Button( self, wx.ID_ANY, u"Begin", wx.DefaultPosition, wx.DefaultSize, 0 )
        bSizer9.Add( self.m_button1, 1, wx.ALL, 5 )
        bSizer1.Add( bSizer9, 1, wx.EXPAND, 5 )
        self.SetSizer( bSizer1 )
        self.Layout()
        self.Centre( wx.BOTH )
        self.m_button1.Bind( wx.EVT_BUTTON, self.submit )
        
        self.app = app
        self.container = container
    
    def submit(self, event):
        nick, addr, port = [p.GetValue() for p in (self.startnick, self.remoteaddr, self.m_spinCtrl1)]
        
        while len(self.container) > 0:
            self.container.pop()
            
        for _ in range(3):
            self.container.append(0)
            
        self.container[0] = nick
        self.container[1] = addr
        self.container[2] = port
        
        self.app.post_dialog()
        self.Close(True)

class MainWindow(wx.Frame):
    def __init__(self, parent, app):
        # massive blob of layout code >_>
        # DO NOT TOUCH! 
        self.app = app
        wx.Frame.__init__ ( self, parent, id = wx.ID_ANY, title = wx.EmptyString, pos = wx.DefaultPosition, size = wx.Size( 1007,568 ), style = wx.DEFAULT_FRAME_STYLE|wx.TAB_TRAVERSAL )
        self.SetSizeHints( wx.DefaultSize, wx.DefaultSize )
        bSizer1 = wx.BoxSizer( wx.HORIZONTAL )
        bSizer2 = wx.BoxSizer( wx.VERTICAL )
        bSizer4 = wx.BoxSizer( wx.HORIZONTAL )
        self.ipalabel = wx.StaticText( self, wx.ID_ANY, "IP Address", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.ipalabel.Wrap( -1 )
        bSizer4.Add( self.ipalabel, 0, wx.ALL, 5 )
        self.ipCtrl = wx.TextCtrl( self, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, 0 )
        bSizer4.Add( self.ipCtrl, 0, wx.ALL, 5 )
        bSizer2.Add( bSizer4, 0, wx.EXPAND, 5 )
        bSizer5 = wx.BoxSizer( wx.HORIZONTAL )
        self.m_button1 = wx.Button( self, wx.ID_ANY, "Connect!", wx.DefaultPosition, wx.DefaultSize, 0 )
        bSizer5.Add( self.m_button1, 10, wx.ALL, 5 )
        bSizer2.Add( bSizer5, 0, wx.EXPAND, 5 )
        self.m_staticline2 = wx.StaticLine( self, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, wx.LI_HORIZONTAL )
        bSizer2.Add( self.m_staticline2, 0, wx.EXPAND |wx.ALL, 5 )
        bSizer6 = wx.BoxSizer( wx.HORIZONTAL )
        self.m_staticText2 = wx.StaticText( self, wx.ID_ANY, "Nickname", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_staticText2.Wrap( -1 )
        bSizer6.Add( self.m_staticText2, 0, wx.ALL, 5 )
        self.nickCtrl = wx.TextCtrl( self, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, 0 )
        bSizer6.Add( self.nickCtrl, 0, wx.ALL, 5 )
        bSizer2.Add( bSizer6, 0, wx.EXPAND, 5 )
        bSizer8 = wx.BoxSizer( wx.HORIZONTAL )
        self.m_button2 = wx.Button( self, wx.ID_ANY, "Set Nickname", wx.DefaultPosition, wx.DefaultSize, 0 )
        bSizer8.Add( self.m_button2, 1, wx.ALL, 5 )
        bSizer2.Add( bSizer8, 0, wx.EXPAND, 5 )
        bSizer10 = wx.BoxSizer( wx.VERTICAL )
        self.m_staticText3 = wx.StaticText( self, wx.ID_ANY, "Users", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_staticText3.Wrap( -1 )
        bSizer10.Add( self.m_staticText3, 0, wx.ALL, 5 )
        bSizer11 = wx.BoxSizer( wx.HORIZONTAL )
        self.userList = wx.TextCtrl( self, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, wx.TE_DONTWRAP|wx.TE_MULTILINE|wx.TE_READONLY )
        bSizer11.Add( self.userList, 1, wx.ALL|wx.EXPAND, 5 )
        bSizer10.Add( bSizer11, 1, wx.EXPAND, 5 )
        bSizer2.Add( bSizer10, 1, wx.EXPAND, 5 )
        self.m_staticline3 = wx.StaticLine( self, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, wx.LI_HORIZONTAL )
        bSizer2.Add( self.m_staticline3, 0, wx.EXPAND |wx.ALL, 5 )
        bSizer9 = wx.BoxSizer( wx.HORIZONTAL )
        self.m_button3 = wx.Button( self, wx.ID_ANY, "Quit", wx.DefaultPosition, wx.DefaultSize, 0 )
        bSizer9.Add( self.m_button3, 1, wx.ALL, 5 )
        bSizer2.Add( bSizer9, 0, wx.EXPAND, 5 )
        bSizer1.Add( bSizer2, 0, wx.EXPAND, 5 )
        self.m_staticline4 = wx.StaticLine( self, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, wx.LI_HORIZONTAL )
        bSizer1.Add( self.m_staticline4, 0, wx.EXPAND |wx.ALL, 5 )
        bSizer3 = wx.BoxSizer( wx.VERTICAL )
        self.buffer = wx.TextCtrl( self, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, wx.TE_MULTILINE|wx.TE_READONLY )
        bSizer3.Add( self.buffer, 1, wx.ALL|wx.EXPAND, 5 )
        bSizer12 = wx.BoxSizer( wx.HORIZONTAL )
        self.mynick = wx.StaticText( self, wx.ID_ANY, app.nickname, wx.DefaultPosition, wx.DefaultSize, 0 )
        self.mynick.Wrap( -1 )
        bSizer12.Add( self.mynick, 0, wx.ALL, 5 )
        self.mymsg = wx.TextCtrl( self, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, wx.TE_PROCESS_ENTER )
        bSizer12.Add( self.mymsg, 1, wx.ALL, 5 )
        bSizer3.Add( bSizer12, 0, wx.EXPAND, 5 )
        bSizer1.Add( bSizer3, 1, wx.EXPAND, 5 )
        self.SetSizer( bSizer1 )
        self.Layout()
        self.Centre( wx.BOTH )
        
        # Events
        self.m_button1.Bind( wx.EVT_BUTTON, self.connect )
        self.m_button2.Bind( wx.EVT_BUTTON, self.set_nick )
        self.m_button3.Bind( wx.EVT_BUTTON, self.quit )
        self.mymsg.Bind( wx.EVT_TEXT_ENTER, self.send_message )
        
        self.Show(True)
    
    def set_nick(self, event):
        nick = self.nickCtrl.GetValue()
        
        if nick == '':
            return
            
        self.mynick.SetLabel(nick)
        self.app.set_nick(nick)
        
        event.Skip()
    
    def connect(self, event):
        if self.app.server is None:
            event.Skip()
            return
    
        addr = self.ipCtrl.GetValue().split(':')
        print("Connecting to: {}".format(addr))
        
        try:
            addr[1] = int(addr[1])
            
        except BaseException as err:
            event.Skip()
            return
            
        self.app.server.connect(tuple(addr))
        
        event.Skip()
    
    def quit(self, event):
        self.app.quit()
        self.Close(True)
        
        event.Skip()
        
    def send_message(self, event):
        msg = self.mymsg.GetValue()
        self.mymsg.SetValue("")
        
        self.app.send_message(msg)
        
        event.Skip()

class WxChatApp(pt.ChatApp):
    def __init__(self):
        super().__init__("")
        print("Starting session.")
        
        self.app = wx.App(False)
        self.frame = MainWindow(None, self)
        self.pos = 0
        
    def start(self):
        print("Initialization dialog:")
        self.initial = []
        dialog = InitDialog(None, self.initial, self)
        dialog.ShowModal()
        self.app.MainLoop()
        
    def post_dialog(self):
        nick, addr, port = self.initial
        eport = None
        
        if ":" in addr:
            addr = addr.split(':')
            eport = int(addr[1])
            addr = addr[0]
        
        print("Starting peer...")
        server = pb.PeerServer(addr, port=port)
        server.register_app(self)
        server.start_loop()
        self.frame.mynick.SetLabel(nick)
        self.set_nick(nick)
        
    def on_app_message(self):
        while self.pos < len(self.messages) - 1:
            self.pos += 1
            
            m = self.messages[self.pos]
            mtype = m[0]
            
            if mtype == "MESSAGE":
                self.frame.buffer.SetValue(self.frame.buffer.GetValue() + "\n<{}> {}".format(self.nicknames[m[2]], m[4]))
                
            elif mtype == "QUIT":
                self.frame.buffer.SetValue(self.frame.buffer.GetValue() + "\n--- {} has quit".format(self.nicknames[m[3]]))
                
            elif mtype == "JOIN":
                self.frame.buffer.SetValue(self.frame.buffer.GetValue() + "\n+++ {} has joined".format(self.nicknames[m[3]]))
        
def main():
    app = WxChatApp()
    app.start()
    
if __name__ == "__main__":
    main()